from __future__ import annotations

from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, cast

from numpy import empty, ndarray
from pandas import Index
from typing_extensions import override
from xarray import DataArray
from xarray.core.types import ErrorOptionsWithWarn
from zarr import Array, suppress

from utilities.numpy import NDArray1
from utilities.pathlib import PathLike
from utilities.sentinel import sentinel
from utilities.text import ensure_str
from utilities.zarr import (
    InvalidDimensionError,
    IselIndexer,
    NDArrayWithIndexes,
    yield_group_and_array,
)

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import IndexA


def save_data_array_to_disk(
    array: DataArray,
    path: PathLike,
    /,
    *,
    overwrite: bool = False,
    chunks: bool | int | tuple[int | None, ...] = True,
) -> None:
    """Save a `DataArray` to disk."""
    name_use = None if (name := array.name) is None else ensure_str(name)
    with yield_data_array_on_disk(
        cast(Mapping[str, Any], array.coords),
        path,
        overwrite=overwrite,
        dtype=array.dtype,
        chunks=chunks,
        name=name_use,
    ) as z_array:
        if (values := array.to_numpy()).shape == ():
            z_array[:] = values.item()
        else:
            z_array[:] = values


@contextmanager
def yield_data_array_on_disk(
    coords: Mapping[str, Any],
    path: PathLike,
    /,
    *,
    overwrite: bool = False,
    dtype: Any = float,
    fill_value: Any = sentinel,
    chunks: bool | int | tuple[int | None, ...] = True,
    name: str | None = None,
) -> Iterator[Array]:
    """Save a `DataArray`, yielding a view into its values."""
    indexes: dict[str, NDArray1] = {}
    for coord, value in coords.items():
        with suppress(NotOneDimensionalArrayError):
            indexes[coord] = _to_ndarray1(value)
    with yield_group_and_array(
        indexes,
        path,
        overwrite=overwrite,
        dtype=dtype,
        fill_value=fill_value,
        chunks=chunks,
    ) as (root, array):
        root.attrs["coords"] = tuple(coords)
        for coord, value in coords.items():
            if coord not in indexes:
                root.attrs[f"coord_{coord}"] = value.item()
        root.attrs["name"] = name
        yield array


def _to_ndarray1(x: Any, /) -> NDArray1:
    """Convert a coordinate into a 1-dimensional array."""
    if isinstance(x, ndarray):
        if x.ndim == 1:
            return x
        msg = f"{x=}"
        raise NotOneDimensionalArrayError(msg)
    if isinstance(x, DataArray | Index):
        if x.ndim == 1:
            return x.to_numpy()
        msg = f"{x=}"
        raise NotOneDimensionalArrayError(msg)
    msg = f"{x=}"
    raise NotOneDimensionalArrayError(msg)


class NotOneDimensionalArrayError(ValueError):
    """Raised when an object is not a 1-dimensional array."""


class DataArrayOnDisk(NDArrayWithIndexes):
    """A `DataArray` stored on disk."""

    @property
    def coords(self) -> dict[str, Any]:
        """The coordinates of the underlying array."""
        return {coord: self._get_coord(coord) for coord in self.attrs["coords"]}

    @property
    def da(self) -> DataArray:
        """Alias for `data_array`."""
        return self.data_array

    @property
    def data_array(self) -> DataArray:
        """The underlying `DataArray`."""
        return DataArray(self.ndarray, self.coords, self.dims, self.name)

    @property
    @override
    def indexes(self) -> dict[str, IndexA]:  # type: ignore
        """The indexes of the underlying array."""
        return {ensure_str(dim): Index(index) for dim, index in super().indexes.items()}

    @override
    def isel(
        self,
        indexers: Mapping[str, IselIndexer] | None = None,
        /,
        *,
        drop: bool = False,
        missing_dims: ErrorOptionsWithWarn = "raise",
        **indexer_kwargs: IselIndexer,
    ) -> DataArray:
        """Select orthogonally using integer indexes."""
        empty = self._empty.isel(
            indexers, drop=drop, missing_dims=missing_dims, **indexer_kwargs
        )
        return DataArray(
            super().isel(indexers, **indexer_kwargs),
            empty.coords,
            empty.dims,
            self.name,
        )

    @property
    def name(self) -> str | None:
        """The name of the underlying array."""
        return self.attrs["name"]

    @override
    def sel(
        self,
        indexers: Mapping[str, Any] | None = None,
        /,
        *,
        method: str | None = None,
        tolerance: Any = None,
        drop: bool = False,
        **indexer_kwargs: Any,
    ) -> DataArray:
        """Select orthogonally using index values."""
        empty = self._empty.sel(
            indexers, method=method, tolerance=tolerance, drop=drop, **indexer_kwargs
        )
        return DataArray(
            super().sel(indexers, **indexer_kwargs), empty.coords, empty.dims, self.name
        )

    @property
    def _empty(self, /) -> DataArray:
        """An empty DataArray, for slicing."""
        return DataArray(
            empty(self.shape, dtype=bool), self.coords, self.dims, self.name
        )

    def _get_coord(self, coord: str, /) -> Any:
        """Get a coordinate by name."""
        try:
            return self._get_index_by_name(coord)
        except InvalidDimensionError:
            return self.attrs[f"coord_{coord}"]


__all__ = [
    "DataArrayOnDisk",
    "NotOneDimensionalArrayError",
    "save_data_array_to_disk",
    "yield_data_array_on_disk",
]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject, data, dictionaries, floats, integers
from numpy import arange, array, isclose, nan, sort, zeros
from numpy.testing import assert_equal
from pytest import mark, param, raises
from zarr import open_array
from zarr.errors import BoundsCheckError

from utilities.class_name import get_class_name
from utilities.hypothesis import float_arrays, int_arrays, temp_paths, text_ascii
from utilities.numpy import NDArrayI1, NDArrayO1, datetime64D, datetime64ns
from utilities.zarr import (
    InvalidIndexValueError,
    NDArrayWithIndexes,
    ffill_non_nan_slices,
    yield_array_with_indexes,
)

indexes1d = int_arrays(shape=integers(0, 10), unique=True).map(sort)


class TestFFillNonNanSlices:
    def test_main(self, *, tmp_path: Path) -> None:
        arr = array(
            [[0.1, nan, nan, 0.2], 4 * [nan], [0.3, nan, nan, nan]], dtype=float
        )
        z_arr = open_array(tmp_path.joinpath("array"), shape=arr.shape, dtype=float)
        z_arr[:] = arr
        ffill_non_nan_slices(z_arr)
        expected = array(
            [[0.1, 0.1, 0.1, 0.2], 4 * [nan], [0.3, 0.3, 0.3, nan]], dtype=float
        )
        assert_equal(z_arr[:], expected)


class TestNDArrayWithIndexes:
    @given(
        data=data(),
        indexes=dictionaries(text_ascii(), indexes1d, max_size=3),
        root=temp_paths(),
    )
    def test_main(
        self, *, data: DataObject, indexes: Mapping[str, NDArrayI1], root: Path
    ) -> None:
        shape = tuple(map(len, indexes.values()))
        arrays = float_arrays(shape=shape, allow_nan=True, allow_inf=True)
        if shape == ():
            arrays |= floats(allow_nan=True, allow_infinity=True).map(array)
        arr = data.draw(arrays)
        path = root.joinpath("array")
        with yield_array_with_indexes(indexes, path) as z_array:
            if shape == ():
                z_array[:] = arr.item()
                exp_size = 0
                is_scalar = True
            else:
                z_array[:] = arr
                exp_size = arr.size
                is_scalar = False
        view = NDArrayWithIndexes(path)
        assert view.dims == tuple(indexes)
        assert view.dtype == float
        assert set(view.indexes) == set(indexes)
        for dim, index in view.indexes.items():
            assert_equal(index, indexes[dim])
        assert view.is_scalar is is_scalar
        assert view.is_non_scalar is (not is_scalar)
        assert_equal(view.ndarray, arr)
        assert view.ndim == len(indexes)
        assert view.shape == shape
        assert view.size == exp_size
        assert view.sizes == dict(zip(indexes, shape, strict=True))

    @given(root=temp_paths())
    def test_dtype(self, *, root: Path) -> None:
        path = root.joinpath("array")
        with yield_array_with_indexes({}, path, dtype=int):
            pass
        view = NDArrayWithIndexes(path)
        assert view.dtype == int

    @given(
        indexes=dictionaries(text_ascii(), indexes1d, max_size=3),
        root=temp_paths(),
        fill_value=floats(allow_infinity=True, allow_nan=True),
    )
    def test_fill_value(
        self, *, indexes: Mapping[str, NDArrayI1], root: Path, fill_value: float
    ) -> None:
        path = root.joinpath("array")
        with yield_array_with_indexes(indexes, path, fill_value=fill_value):
            pass
        view = NDArrayWithIndexes(path)
        assert isclose(view.ndarray, fill_value, equal_nan=True).all()

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param({"x": 0}, array([0, 1, 2])),
            param({"x": -1}, array([3, 4, 5])),
            param({"x": slice(None, 1)}, array([[0, 1, 2]])),
            param({"x": []}, zeros((0, 3), dtype=int)),
            param({"x": array([True, False])}, array([[0, 1, 2]])),
            param({"x": array([0])}, array([[0, 1, 2]])),
            param({"x": 0, "y": 0}, 0),
            param({"x": 0, "y": -1}, 2),
            param({"x": 0, "y": slice(None, 1)}, array([0])),
            param({"x": 0, "y": []}, array([])),
            param({"x": 0, "y": array([True, False, False])}, array([0])),
            param({"x": 0, "y": array([0])}, array([0])),
        ],
    )
    def test_isel(
        self, *, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        indexes = {"x": arange(2), "y": arange(3)}
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(6, dtype=int).reshape(2, 3)
        view = NDArrayWithIndexes(path)
        assert_equal(view.isel(indexer), expected)

    @mark.parametrize("indexer", [param({"x": 2}), param({"x": [2]})])
    def test_isel_error(self, *, tmp_path: Path, indexer: Mapping[str, Any]) -> None:
        indexes = {"x": arange(2), "y": arange(3)}
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(6, dtype=int).reshape(2, 3)
        view = NDArrayWithIndexes(path)
        with raises(BoundsCheckError):
            _ = view.isel(indexer)

    @mark.parametrize("func", [param(repr), param(str)])
    def test_repr_and_str(self, *, func: Callable[[Any], str], tmp_path: Path) -> None:
        view = NDArrayWithIndexes(tmp_path)
        cls = get_class_name(NDArrayWithIndexes)
        assert func(view) == f"{cls}({tmp_path})"

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param({"x": "x0"}, array([0, 1, 2])),
            param({"x": []}, zeros((0, 3), dtype=int)),
            param({"x": ["x0"]}, array([[0, 1, 2]])),
            param({"x": ["x0", "x1"]}, array([[0, 1, 2], [3, 4, 5]])),
            param({"x": "x0", "y": "y0"}, 0),
            param({"x": "x0", "y": []}, zeros(0, dtype=int)),
            param({"x": "x0", "y": ["y0"]}, array([0])),
            param({"x": "x0", "y": ["y0", "y1"]}, array([0, 1])),
        ],
    )
    def test_sel(
        self, *, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        indexes = {"x": array(["x0", "x1"]), "y": array(["y0", "y1", "y2"])}
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(6, dtype=int).reshape(2, 3)
        view = NDArrayWithIndexes(path)
        assert_equal(view.sel(indexer), expected)

    @mark.parametrize(
        ("index", "indexer"),
        [
            param(array(["x0", "x1"]), {"x": "x2"}),
            param(array(["x0", "x0"]), {"x": "x0"}),
            param(array(["x0", "x0"]), {"x": ["x0"]}),
        ],
    )
    def test_sel_error(
        self, *, tmp_path: Path, index: NDArrayO1, indexer: Mapping[str, Any]
    ) -> None:
        indexes = {"x": index}
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(2, dtype=int)
        view = NDArrayWithIndexes(path)
        with raises(InvalidIndexValueError):
            _ = view.sel(indexer)

    def test_missing(self, *, tmp_path: Path) -> None:
        with raises(FileNotFoundError):
            _ = NDArrayWithIndexes(tmp_path.joinpath("array"))

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param({"x": dt.date(2000, 1, 1)}, 0),
            param({"x": "2000-01-01"}, 0),
            param({"x": [dt.date(2000, 1, 1)]}, array([0])),
            param({"x": ["2000-01-01"]}, array([0])),
            param({"x": [dt.date(2000, 1, 1), dt.date(2000, 1, 2)]}, array([0, 1])),
            param({"x": [dt.date(2000, 1, 1), "2000-01-02"]}, array([0, 1])),
            param({"x": ["2000-01-01", dt.date(2000, 1, 2)]}, array([0, 1])),
            param({"x": ["2000-01-01", "2000-01-02"]}, array([0, 1])),
        ],
    )
    def test_sel_with_date(
        self, *, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        indexes = {
            "x": array([dt.date(2000, 1, i) for i in range(1, 4)], dtype=datetime64D)
        }
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(3)
        view = NDArrayWithIndexes(path)
        assert_equal(view.sel(indexer), expected)

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param({"x": dt.datetime(2000, 1, 1)}, 0),  # noqa: DTZ001
            param({"x": "2000-01-01"}, 0),
            param({"x": [dt.datetime(2000, 1, 1)]}, array([0])),  # noqa: DTZ001
            param({"x": ["2000-01-01"]}, array([0])),
            param(
                {
                    "x": [
                        dt.datetime(2000, 1, 1),  # noqa: DTZ001
                        dt.datetime(2000, 1, 2),  # noqa: DTZ001
                    ]
                },
                array([0, 1]),
            ),
            param(
                {"x": [dt.datetime(2000, 1, 1), "2000-01-02"]},  # noqa: DTZ001
                array([0, 1]),
            ),
            param(
                {"x": ["2000-01-01", dt.datetime(2000, 1, 2)]},  # noqa: DTZ001
                array([0, 1]),
            ),
            param({"x": ["2000-01-01", "2000-01-02"]}, array([0, 1])),
        ],
    )
    def test_sel_with_datetime(
        self, *, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        indexes = {
            "x": array([dt.date(2000, 1, i) for i in range(1, 4)], dtype=datetime64ns)
        }
        path = tmp_path.joinpath("array")
        with yield_array_with_indexes(indexes, path, dtype=int) as z_array:
            z_array[:] = arange(3, dtype=int)
        view = NDArrayWithIndexes(path)
        assert_equal(view.sel(indexer), expected)



from __future__ import annotations

from pathlib import Path
from subprocess import run

from hypothesis.strategies import DrawFn, composite

from utilities.hypothesis.hypothesis import MaybeSearchStrategy, lift_draw, temp_paths
from utilities.pathlib import temp_cwd


@composite
def git_repos(
    _draw: DrawFn, /, *, branch: MaybeSearchStrategy[str | None] = None
) -> Path:
    draw = lift_draw(_draw)
    path = draw(temp_paths())
    with temp_cwd(path):
        _ = run(["git", "init"], check=True)  # noqa: S603, S607
        _ = run(
            ["git", "config", "user.name", "User"],  # noqa: S603, S607
            check=True,
        )
        _ = run(
            ["git", "config", "user.email", "a@z.com"],  # noqa: S603, S607
            check=True,
        )
        file = path.joinpath("file")
        file.touch()
        file_str = str(file)
        _ = run(["git", "add", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "add"], check=True)  # noqa: S603, S607
        _ = run(["git", "rm", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "rm"], check=True)  # noqa: S603, S607
        if (branch := draw(branch)) is not None:
            _ = run(
                ["git", "checkout", "-b", branch],  # noqa: S603, S607
                check=True,
            )
    return path


__all__ = ["git_repos"]



from __future__ import annotations

import datetime as dt
from collections.abc import Iterable, Iterator
from datetime import timezone
from decimal import Decimal
from functools import partial
from typing import Any, cast, overload

from numpy import int64
from pandas import DataFrame, DatetimeTZDtype, Series, concat
from sqlalchemy import Column, Connection, Engine, Row, Select, Table
from sqlalchemy.sql.base import ReadOnlyColumnCollection

from utilities.datetime import UTC
from utilities.numpy import datetime64ns
from utilities.pandas import (
    Int64,
    astype,
    boolean,
    datetime64nsutc,
    string,
    timestamp_to_date,
    timestamp_to_datetime,
)
from utilities.sqlalchemy import (
    CHUNK_SIZE_FRAC,
    check_dataframe_schema_against_table,
    check_selectable_for_duplicate_columns,
    insert_items,
    yield_connection,
)
from utilities.text import snake_case_mappings


def insert_pandas_dataframe(
    engine_or_conn: Engine | Connection,
    df: DataFrame,
    table_or_mapped_class: Table | type[Any],
    /,
    *,
    allow_naive_datetimes: bool = False,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> None:
    """Insert a DataFrame into a database."""
    check_dtype = partial(
        _check_pandas_series, allow_naive_datetimes=allow_naive_datetimes
    )
    mapping = check_dataframe_schema_against_table(
        dict(df.dtypes), table_or_mapped_class, check_dtype, snake=snake
    )
    df_items = df[list(mapping)].rename(columns=mapping)
    for name, col in df_items.items():
        col_obj = col.astype(object).where(col.notna(), None)
        if col.dtype == datetime64ns:
            is_not_null = col.notnull()
            values = col.loc[is_not_null]
            if (values == values.dt.normalize()).all():
                func = timestamp_to_date
            else:
                func = timestamp_to_datetime
            col_obj.loc[is_not_null] = values.map(func)
        df_items[name] = col_obj
    items = df_items.to_dict(orient="records")
    if (len(df) > 0) and (len(items) == 0):
        msg = f"{df=}, {items=}"
        raise PandasDataFrameYieldsNoRowsError(msg)
    return insert_items(
        engine_or_conn, (items, table_or_mapped_class), chunk_size_frac=chunk_size_frac
    )


class PandasDataFrameYieldsNoRowsError(ValueError):
    """Raised when a DataFrame yields no rows to insert."""


def _check_pandas_series(
    dtype: Any, py_type: type, /, *, allow_naive_datetimes: bool = False
) -> bool:
    is_bool = (dtype == bool) or (dtype == boolean)  # noqa: PLR1714
    is_int = (dtype == int) or (dtype == int64) or (dtype == Int64)  # noqa: PLR1714
    return (
        (is_bool and issubclass(py_type, bool))
        or ((dtype == datetime64ns) and issubclass(py_type, dt.date))
        or ((dtype == datetime64nsutc) and issubclass(py_type, dt.datetime))
        or (
            allow_naive_datetimes
            and (dtype == datetime64ns)
            and issubclass(py_type, dt.datetime)
        )
        or ((dtype == float) and issubclass(py_type, float))
        or (is_int and issubclass(py_type, int))
        or ((dtype == string) and issubclass(py_type, str))
    )


@overload
def select_to_pandas_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    stream: int,
    time_zone: timezone = ...,
    snake: bool = ...,
) -> Iterator[DataFrame]:
    ...


@overload
def select_to_pandas_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    stream: None = ...,
    time_zone: timezone = ...,
    snake: bool = False,
) -> DataFrame:
    ...


def select_to_pandas_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    stream: int | None = None,
    time_zone: timezone = UTC,
    snake: bool = False,
) -> DataFrame | Iterator[DataFrame]:
    """Read a table from a database into a DataFrame.

    Optionally stream it in chunks.
    """
    check_selectable_for_duplicate_columns(sel)
    if stream is None:
        with yield_connection(engine_or_conn) as conn:
            rows = conn.execute(sel).all()
            return _rows_to_dataframe(sel, rows, time_zone=time_zone, snake=snake)
    return _stream_dataframes(
        sel, engine_or_conn, stream, time_zone=time_zone, snake=snake
    )


def _rows_to_dataframe(
    sel: Select[Any],
    rows: Iterable[Row[Any]],
    /,
    *,
    time_zone: timezone = UTC,
    snake: bool = False,
) -> DataFrame:
    """Convert a set of rows into a DataFrame."""
    columns: ReadOnlyColumnCollection = cast(Any, sel).selected_columns
    dtypes = {
        col.name: _table_column_to_dtype(col, time_zone=time_zone)
        for col in columns.values()
    }
    rows = list(rows)
    if len(rows) >= 1:
        by_cols = zip(*rows, strict=True)
        series = (
            Series(data, dtype=dtype, name=name)
            for data, (name, dtype) in zip(by_cols, dtypes.items(), strict=True)
        )
        df = concat(series, axis=1)
    else:
        df = DataFrame(columns=list(dtypes))
        df = astype(df, dtypes)
    return _dataframe_columns_to_snake(df) if snake else df


def _table_column_to_dtype(column: Column[Any], /, *, time_zone: timezone = UTC) -> Any:
    """Map a table column to a DataFrame dtype."""
    db_type = column.type
    py_type = db_type.python_type
    if issubclass(py_type, bool):
        return boolean
    if issubclass(py_type, dt.date) and not issubclass(py_type, dt.datetime):
        return datetime64ns
    if issubclass(py_type, dt.datetime):
        has_tz: bool = cast(Any, db_type).timezone
        return DatetimeTZDtype(tz=time_zone) if has_tz else datetime64ns
    if issubclass(py_type, float | Decimal):
        return float
    if issubclass(py_type, int):
        return Int64
    if issubclass(py_type, str):
        return string
    msg = f"{column=}"
    raise ColumnToPandasDTypeError(msg)


class ColumnToPandasDTypeError(TypeError):
    """Raised when an invalid column type is encountered."""


def _dataframe_columns_to_snake(df: DataFrame, /) -> DataFrame:
    """Convert the columns of a DataFrame to snake case."""
    mapping = snake_case_mappings(list(df.columns))
    return df.rename(columns=mapping)


def _stream_dataframes(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    stream: int,
    /,
    *,
    time_zone: timezone = UTC,
    snake: bool = False,
) -> Iterator[DataFrame]:
    if stream <= 0:
        raise NonPositiveStreamError(str(stream))
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield from _stream_dataframes(
                sel, conn, stream, time_zone=time_zone, snake=snake
            )
    else:
        for rows in (
            engine_or_conn.execution_options(yield_per=stream).execute(sel).partitions()
        ):
            yield _rows_to_dataframe(sel, rows, time_zone=time_zone, snake=snake)


class NonPositiveStreamError(ValueError):
    """Raised when the `stream` arguemnt is non-positive."""


__all__ = [
    "ColumnToPandasDTypeError",
    "insert_pandas_dataframe",
    "NonPositiveStreamError",
    "PandasDataFrameYieldsNoRowsError",
    "select_to_pandas_dataframe",
]



from __future__ import annotations

from collections.abc import Iterator
from os import getenv
from typing import TypedDict, cast

import pytest
from _pytest.logging import LogCaptureFixture
from hypothesis import Verbosity, settings
from loguru import logger


class Kwargs(TypedDict, total=False):
    deadline: None
    print_blob: bool
    report_multiple_bugs: bool


kwargs = cast(
    Kwargs, {"deadline": None, "print_blob": True, "report_multiple_bugs": False}
)
settings.register_profile("default", max_examples=100, **kwargs)
settings.register_profile("dev", max_examples=10, **kwargs)
settings.register_profile("ci", max_examples=1000, **kwargs)
settings.register_profile(
    "debug", max_examples=10, verbosity=Verbosity.verbose, **kwargs
)
settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))


@pytest.fixture()
def caplog(*, caplog: pytest.LogCaptureFixture) -> Iterator[LogCaptureFixture]:
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from utilities.tempfile import TEMP_DIR
from utilities.typed_settings import click_field


@dataclass(frozen=True)
class Config:
    """Settings for the `clean_dir` script."""

    paths: set[Path] = click_field(
        default=frozenset([TEMP_DIR]), param_decls=("-p", "--path")
    )
    days: int = click_field(default=7, param_decls=("-d", "--days"))
    chunk_size: Optional[int] = click_field(  # noqa: UP007
        default=None, param_decls=("-cs", "--chunk-size")
    )
    dry_run: bool = click_field(default=False, param_decls=("-dr", "--dry-run"))


@dataclass(frozen=True)
class Item:
    """An item to clean up."""

    path: Path
    clean: Callable[[], None]



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from fpdf import FPDF
from fpdf.enums import XPos, YPos
from typing_extensions import override

from utilities.datetime import get_now, local_timezone
from utilities.holoviews import save_plot
from utilities.tempfile import TemporaryDirectory


class _BasePDF(FPDF):
    """Base class for PDFs."""

    def add_fixed_width_text(self, text: str, /) -> None:
        """Add a block of fixed witth text."""
        self.set_font("Courier")
        _ = self.write(txt=text)
        self.ln()

    def add_plot(self, plot: Any, /) -> None:  # pragma: no cover
        with TemporaryDirectory() as temp:
            path = temp.joinpath("image.png")
            save_plot(plot, path)
            _ = self.image(path, w=self.epw)


@contextmanager
def yield_pdf(*, header: str | None = None) -> Iterator[_BasePDF]:
    """Yield a PDF."""

    class OutputPDF(_BasePDF):
        @override
        def header(self) -> None:
            if header is not None:
                self.set_font(family="Helvetica", style="B", size=15)
                _ = self.cell(w=80)
                _ = self.cell(
                    w=30,
                    h=10,
                    txt=header,
                    border=0,
                    align="C",
                    new_x=XPos.RIGHT,
                    new_y=YPos.TOP,
                )
                self.ln(20)

        @override
        def footer(self) -> None:
            self.set_y(-15)
            self.set_font(family="Helvetica", style="I", size=8)
            page_no, now = self.page_no(), get_now(tz=local_timezone())
            txt = f"page {page_no}/{{nb}}; {now:%Y-%m-%d %H:%M:%S}"
            _ = self.cell(
                w=0,
                h=10,
                txt=txt,
                border=0,
                align="C",
                new_x=XPos.RIGHT,
                new_y=YPos.TOP,
            )

    pdf = OutputPDF(orientation="portrait", unit="mm", format="A4")
    pdf.alias_nb_pages()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    yield pdf


__all__ = ["yield_pdf"]



from __future__ import annotations

from hashlib import md5
from pickle import dumps
from typing import Any

from utilities.json import JsonSerializationError, serialize


def md5_hash(obj: Any, /) -> str:
    """Compute the MD5 hash of an arbitrary object."""
    try:
        ser = serialize(obj).encode()
    except JsonSerializationError:
        ser = dumps(obj)
    return md5(ser, usedforsecurity=False).hexdigest()


__all__ = ["md5_hash"]



from __future__ import annotations

from utilities.numpy.numpy import (
    DATE_MAX_AS_DATETIME64,
    DATE_MAX_AS_INT,
    DATE_MIN_AS_DATETIME64,
    DATE_MIN_AS_INT,
    DATETIME_MAX_AS_DATETIME64,
    DATETIME_MAX_AS_INT,
    DATETIME_MIN_AS_DATETIME64,
    DATETIME_MIN_AS_INT,
    DateOverflowError,
    Datetime64Kind,
    Datetime64Unit,
    DTypeB,
    DTypeDns,
    DTypeF,
    DTypeI,
    DTypeO,
    EmptyNumpyConcatenateError,
    InfElementsError,
    InvalidDTypeError,
    IsFinite,
    IsFiniteAndIntegral,
    IsFiniteAndIntegralOrNan,
    IsFiniteAndNegative,
    IsFiniteAndNegativeOrNan,
    IsFiniteAndNonNegative,
    IsFiniteAndNonNegativeOrNan,
    IsFiniteAndNonPositive,
    IsFiniteAndNonPositiveOrNan,
    IsFiniteAndNonZero,
    IsFiniteAndNonZeroOrNan,
    IsFiniteAndPositive,
    IsFiniteAndPositiveOrNan,
    IsFiniteOrNan,
    IsIntegral,
    IsIntegralOrNan,
    IsNegative,
    IsNegativeOrNan,
    IsNonNegative,
    IsNonNegativeOrNan,
    IsNonPositive,
    IsNonPositiveOrNan,
    IsNonZero,
    IsNonZeroOrNan,
    IsPositive,
    IsPositiveOrNan,
    IsZero,
    IsZeroOrFiniteAndNonMicro,
    IsZeroOrFiniteAndNonMicroOrNan,
    IsZeroOrNan,
    IsZeroOrNonMicro,
    IsZeroOrNonMicroOrNan,
    LossOfNanosecondsError,
    MultipleTrueElementsError,
    NanElementsError,
    NDArray0,
    NDArray1,
    NDArray2,
    NDArray3,
    NDArrayA,
    NDArrayB,
    NDArrayB0,
    NDArrayB1,
    NDArrayB2,
    NDArrayB3,
    NDArrayD,
    NDArrayD0,
    NDArrayD1,
    NDArrayD2,
    NDArrayD3,
    NDArrayDas,
    NDArrayDas0,
    NDArrayDas1,
    NDArrayDas2,
    NDArrayDas3,
    NDArrayDD,
    NDArrayDD0,
    NDArrayDD1,
    NDArrayDD2,
    NDArrayDD3,
    NDArrayDfs,
    NDArrayDfs0,
    NDArrayDfs1,
    NDArrayDfs2,
    NDArrayDfs3,
    NDArrayDh,
    NDArrayDh0,
    NDArrayDh1,
    NDArrayDh2,
    NDArrayDh3,
    NDArrayDM,
    NDArrayDm,
    NDArrayDM0,
    NDArrayDm0,
    NDArrayDM1,
    NDArrayDm1,
    NDArrayDM2,
    NDArrayDm2,
    NDArrayDM3,
    NDArrayDm3,
    NDArrayDms,
    NDArrayDms0,
    NDArrayDms1,
    NDArrayDms2,
    NDArrayDms3,
    NDArrayDns,
    NDArrayDns0,
    NDArrayDns1,
    NDArrayDns2,
    NDArrayDns3,
    NDArrayDps,
    NDArrayDps0,
    NDArrayDps1,
    NDArrayDps2,
    NDArrayDps3,
    NDArrayDs,
    NDArrayDs0,
    NDArrayDs1,
    NDArrayDs2,
    NDArrayDs3,
    NDArrayDus,
    NDArrayDus0,
    NDArrayDus1,
    NDArrayDus2,
    NDArrayDus3,
    NDArrayDW,
    NDArrayDW0,
    NDArrayDW1,
    NDArrayDW2,
    NDArrayDW3,
    NDArrayDY,
    NDArrayDY0,
    NDArrayDY1,
    NDArrayDY2,
    NDArrayDY3,
    NDArrayF,
    NDArrayF0,
    NDArrayF0Fin,
    NDArrayF0FinInt,
    NDArrayF0FinIntNan,
    NDArrayF0FinNan,
    NDArrayF0FinNeg,
    NDArrayF0FinNegNan,
    NDArrayF0FinNonNeg,
    NDArrayF0FinNonNegNan,
    NDArrayF0FinNonPos,
    NDArrayF0FinNonPosNan,
    NDArrayF0FinNonZr,
    NDArrayF0FinNonZrNan,
    NDArrayF0FinPos,
    NDArrayF0FinPosNan,
    NDArrayF0Int,
    NDArrayF0IntNan,
    NDArrayF0Neg,
    NDArrayF0NegNan,
    NDArrayF0NonNeg,
    NDArrayF0NonNegNan,
    NDArrayF0NonPos,
    NDArrayF0NonPosNan,
    NDArrayF0NonZr,
    NDArrayF0NonZrNan,
    NDArrayF0Pos,
    NDArrayF0PosNan,
    NDArrayF0Zr,
    NDArrayF0ZrFinNonMic,
    NDArrayF0ZrFinNonMicNan,
    NDArrayF0ZrNan,
    NDArrayF0ZrNonMic,
    NDArrayF0ZrNonMicNan,
    NDArrayF1,
    NDArrayF1Fin,
    NDArrayF1FinInt,
    NDArrayF1FinIntNan,
    NDArrayF1FinNan,
    NDArrayF1FinNeg,
    NDArrayF1FinNegNan,
    NDArrayF1FinNonNeg,
    NDArrayF1FinNonNegNan,
    NDArrayF1FinNonPos,
    NDArrayF1FinNonPosNan,
    NDArrayF1FinNonZr,
    NDArrayF1FinNonZrNan,
    NDArrayF1FinPos,
    NDArrayF1FinPosNan,
    NDArrayF1Int,
    NDArrayF1IntNan,
    NDArrayF1Neg,
    NDArrayF1NegNan,
    NDArrayF1NonNeg,
    NDArrayF1NonNegNan,
    NDArrayF1NonPos,
    NDArrayF1NonPosNan,
    NDArrayF1NonZr,
    NDArrayF1NonZrNan,
    NDArrayF1Pos,
    NDArrayF1PosNan,
    NDArrayF1Zr,
    NDArrayF1ZrFinNonMic,
    NDArrayF1ZrFinNonMicNan,
    NDArrayF1ZrNan,
    NDArrayF1ZrNonMic,
    NDArrayF1ZrNonMicNan,
    NDArrayF2,
    NDArrayF2Fin,
    NDArrayF2FinInt,
    NDArrayF2FinIntNan,
    NDArrayF2FinNan,
    NDArrayF2FinNeg,
    NDArrayF2FinNegNan,
    NDArrayF2FinNonNeg,
    NDArrayF2FinNonNegNan,
    NDArrayF2FinNonPos,
    NDArrayF2FinNonPosNan,
    NDArrayF2FinNonZr,
    NDArrayF2FinNonZrNan,
    NDArrayF2FinPos,
    NDArrayF2FinPosNan,
    NDArrayF2Int,
    NDArrayF2IntNan,
    NDArrayF2Neg,
    NDArrayF2NegNan,
    NDArrayF2NonNeg,
    NDArrayF2NonNegNan,
    NDArrayF2NonPos,
    NDArrayF2NonPosNan,
    NDArrayF2NonZr,
    NDArrayF2NonZrNan,
    NDArrayF2Pos,
    NDArrayF2PosNan,
    NDArrayF2Zr,
    NDArrayF2ZrFinNonMic,
    NDArrayF2ZrFinNonMicNan,
    NDArrayF2ZrNan,
    NDArrayF2ZrNonMic,
    NDArrayF2ZrNonMicNan,
    NDArrayF3,
    NDArrayF3Fin,
    NDArrayF3FinInt,
    NDArrayF3FinIntNan,
    NDArrayF3FinNan,
    NDArrayF3FinNeg,
    NDArrayF3FinNegNan,
    NDArrayF3FinNonNeg,
    NDArrayF3FinNonNegNan,
    NDArrayF3FinNonPos,
    NDArrayF3FinNonPosNan,
    NDArrayF3FinNonZr,
    NDArrayF3FinNonZrNan,
    NDArrayF3FinPos,
    NDArrayF3FinPosNan,
    NDArrayF3Int,
    NDArrayF3IntNan,
    NDArrayF3Neg,
    NDArrayF3NegNan,
    NDArrayF3NonNeg,
    NDArrayF3NonNegNan,
    NDArrayF3NonPos,
    NDArrayF3NonPosNan,
    NDArrayF3NonZr,
    NDArrayF3NonZrNan,
    NDArrayF3Pos,
    NDArrayF3PosNan,
    NDArrayF3Zr,
    NDArrayF3ZrFinNonMic,
    NDArrayF3ZrFinNonMicNan,
    NDArrayF3ZrNan,
    NDArrayF3ZrNonMic,
    NDArrayF3ZrNonMicNan,
    NDArrayFFin,
    NDArrayFFinInt,
    NDArrayFFinIntNan,
    NDArrayFFinNan,
    NDArrayFFinNeg,
    NDArrayFFinNegNan,
    NDArrayFFinNonNeg,
    NDArrayFFinNonNegNan,
    NDArrayFFinNonPos,
    NDArrayFFinNonPosNan,
    NDArrayFFinNonZr,
    NDArrayFFinNonZrNan,
    NDArrayFFinPos,
    NDArrayFFinPosNan,
    NDArrayFInt,
    NDArrayFIntNan,
    NDArrayFNeg,
    NDArrayFNegNan,
    NDArrayFNonNeg,
    NDArrayFNonNegNan,
    NDArrayFNonPos,
    NDArrayFNonPosNan,
    NDArrayFNonZr,
    NDArrayFNonZrNan,
    NDArrayFPos,
    NDArrayFPosNan,
    NDArrayFZr,
    NDArrayFZrFinNonMic,
    NDArrayFZrFinNonMicNan,
    NDArrayFZrNan,
    NDArrayFZrNonMic,
    NDArrayFZrNonMicNan,
    NDArrayI,
    NDArrayI0,
    NDArrayI0Neg,
    NDArrayI0NonNeg,
    NDArrayI0NonPos,
    NDArrayI0NonZr,
    NDArrayI0Pos,
    NDArrayI0Zr,
    NDArrayI1,
    NDArrayI1Neg,
    NDArrayI1NonNeg,
    NDArrayI1NonPos,
    NDArrayI1NonZr,
    NDArrayI1Pos,
    NDArrayI1Zr,
    NDArrayI2,
    NDArrayI2Neg,
    NDArrayI2NonNeg,
    NDArrayI2NonPos,
    NDArrayI2NonZr,
    NDArrayI2Pos,
    NDArrayI2Zr,
    NDArrayI3,
    NDArrayI3Neg,
    NDArrayI3NonNeg,
    NDArrayI3NonPos,
    NDArrayI3NonZr,
    NDArrayI3Pos,
    NDArrayI3Zr,
    NDArrayINeg,
    NDArrayINonNeg,
    NDArrayINonPos,
    NDArrayINonZr,
    NDArrayIPos,
    NDArrayIZr,
    NDArrayO,
    NDArrayO0,
    NDArrayO1,
    NDArrayO2,
    NDArrayO3,
    NDim0,
    NDim1,
    NDim2,
    NDim3,
    NonIntegralElementsError,
    NoTrueElementsError,
    ZeroShiftError,
    array_indexer,
    as_int,
    date_to_datetime64,
    datetime64_dtype_to_unit,
    datetime64_to_date,
    datetime64_to_datetime,
    datetime64_to_int,
    datetime64_unit_to_dtype,
    datetime64_unit_to_kind,
    datetime64as,
    datetime64D,
    datetime64fs,
    datetime64h,
    datetime64M,
    datetime64m,
    datetime64ms,
    datetime64ns,
    datetime64ps,
    datetime64s,
    datetime64us,
    datetime64W,
    datetime64Y,
    datetime_to_datetime64,
    discretize,
    ffill_non_nan_slices,
    fillna,
    flatn0,
    get_fill_value,
    has_dtype,
    is_at_least,
    is_at_least_or_nan,
    is_at_most,
    is_at_most_or_nan,
    is_between,
    is_between_or_nan,
    is_empty,
    is_finite_and_integral,
    is_finite_and_integral_or_nan,
    is_finite_and_negative,
    is_finite_and_negative_or_nan,
    is_finite_and_non_negative,
    is_finite_and_non_negative_or_nan,
    is_finite_and_non_positive,
    is_finite_and_non_positive_or_nan,
    is_finite_and_non_zero,
    is_finite_and_non_zero_or_nan,
    is_finite_and_positive,
    is_finite_and_positive_or_nan,
    is_finite_or_nan,
    is_greater_than,
    is_greater_than_or_nan,
    is_integral,
    is_integral_or_nan,
    is_less_than,
    is_less_than_or_nan,
    is_negative,
    is_negative_or_nan,
    is_non_empty,
    is_non_negative,
    is_non_negative_or_nan,
    is_non_positive,
    is_non_positive_or_nan,
    is_non_singular,
    is_non_zero,
    is_non_zero_or_nan,
    is_positive,
    is_positive_or_nan,
    is_positive_semidefinite,
    is_symmetric,
    is_zero,
    is_zero_or_finite_and_non_micro,
    is_zero_or_finite_and_non_micro_or_nan,
    is_zero_or_nan,
    is_zero_or_non_micro,
    is_zero_or_non_micro_or_nan,
    maximum,
    minimum,
    redirect_to_empty_numpy_concatenate_error,
    shift,
    shift_bool,
    year,
)

__all__ = [
    "array_indexer",
    "as_int",
    "DATE_MAX_AS_DATETIME64",
    "DATE_MAX_AS_INT",
    "DATE_MIN_AS_DATETIME64",
    "DATE_MIN_AS_INT",
    "date_to_datetime64",
    "DateOverflowError",
    "DATETIME_MAX_AS_DATETIME64",
    "DATETIME_MAX_AS_INT",
    "DATETIME_MIN_AS_DATETIME64",
    "DATETIME_MIN_AS_INT",
    "datetime_to_datetime64",
    "datetime64_dtype_to_unit",
    "datetime64_to_date",
    "datetime64_to_datetime",
    "datetime64_to_int",
    "datetime64_unit_to_dtype",
    "datetime64_unit_to_kind",
    "datetime64as",
    "datetime64D",
    "datetime64fs",
    "datetime64h",
    "Datetime64Kind",
    "datetime64m",
    "datetime64M",
    "datetime64ms",
    "datetime64ns",
    "datetime64ps",
    "datetime64s",
    "Datetime64Unit",
    "datetime64us",
    "datetime64W",
    "datetime64Y",
    "discretize",
    "DTypeB",
    "DTypeDns",
    "DTypeF",
    "DTypeI",
    "DTypeO",
    "EmptyNumpyConcatenateError",
    "ffill_non_nan_slices",
    "fillna",
    "flatn0",
    "get_fill_value",
    "has_dtype",
    "InfElementsError",
    "InvalidDTypeError",
    "is_at_least_or_nan",
    "is_at_least",
    "is_at_most_or_nan",
    "is_at_most",
    "is_between_or_nan",
    "is_between",
    "is_empty",
    "is_finite_and_integral_or_nan",
    "is_finite_and_integral",
    "is_finite_and_negative_or_nan",
    "is_finite_and_negative",
    "is_finite_and_non_negative_or_nan",
    "is_finite_and_non_negative",
    "is_finite_and_non_positive_or_nan",
    "is_finite_and_non_positive",
    "is_finite_and_non_zero_or_nan",
    "is_finite_and_non_zero",
    "is_finite_and_positive_or_nan",
    "is_finite_and_positive",
    "is_finite_or_nan",
    "is_greater_than_or_nan",
    "is_greater_than",
    "is_integral_or_nan",
    "is_integral",
    "is_less_than_or_nan",
    "is_less_than",
    "is_negative_or_nan",
    "is_negative",
    "is_non_empty",
    "is_non_negative_or_nan",
    "is_non_negative",
    "is_non_positive_or_nan",
    "is_non_positive",
    "is_non_singular",
    "is_non_zero_or_nan",
    "is_non_zero",
    "is_positive_or_nan",
    "is_positive_semidefinite",
    "is_positive",
    "is_symmetric",
    "is_zero_or_finite_and_non_micro_or_nan",
    "is_zero_or_finite_and_non_micro",
    "is_zero_or_nan",
    "is_zero_or_non_micro_or_nan",
    "is_zero_or_non_micro",
    "is_zero",
    "IsFinite",
    "IsFiniteAndIntegral",
    "IsFiniteAndIntegralOrNan",
    "IsFiniteAndNegative",
    "IsFiniteAndNegativeOrNan",
    "IsFiniteAndNonNegative",
    "IsFiniteAndNonNegativeOrNan",
    "IsFiniteAndNonPositive",
    "IsFiniteAndNonPositiveOrNan",
    "IsFiniteAndNonZero",
    "IsFiniteAndNonZeroOrNan",
    "IsFiniteAndPositive",
    "IsFiniteAndPositiveOrNan",
    "IsFiniteOrNan",
    "IsIntegral",
    "IsIntegralOrNan",
    "IsNegative",
    "IsNegativeOrNan",
    "IsNonNegative",
    "IsNonNegativeOrNan",
    "IsNonPositive",
    "IsNonPositiveOrNan",
    "IsNonZero",
    "IsNonZeroOrNan",
    "IsPositive",
    "IsPositiveOrNan",
    "IsZero",
    "IsZeroOrFiniteAndNonMicro",
    "IsZeroOrFiniteAndNonMicroOrNan",
    "IsZeroOrNan",
    "IsZeroOrNonMicro",
    "IsZeroOrNonMicroOrNan",
    "LossOfNanosecondsError",
    "maximum",
    "minimum",
    "MultipleTrueElementsError",
    "NanElementsError",
    "NDArray0",
    "NDArray1",
    "NDArray2",
    "NDArray3",
    "NDArrayA",
    "NDArrayB",
    "NDArrayB0",
    "NDArrayB1",
    "NDArrayB2",
    "NDArrayB3",
    "NDArrayD",
    "NDArrayD0",
    "NDArrayD1",
    "NDArrayD2",
    "NDArrayD3",
    "NDArrayDas",
    "NDArrayDas0",
    "NDArrayDas1",
    "NDArrayDas2",
    "NDArrayDas3",
    "NDArrayDD",
    "NDArrayDD0",
    "NDArrayDD1",
    "NDArrayDD2",
    "NDArrayDD3",
    "NDArrayDfs",
    "NDArrayDfs0",
    "NDArrayDfs1",
    "NDArrayDfs2",
    "NDArrayDfs3",
    "NDArrayDh",
    "NDArrayDh0",
    "NDArrayDh1",
    "NDArrayDh2",
    "NDArrayDh3",
    "NDArrayDm",
    "NDArrayDM",
    "NDArrayDm0",
    "NDArrayDM0",
    "NDArrayDm1",
    "NDArrayDM1",
    "NDArrayDm2",
    "NDArrayDM2",
    "NDArrayDm3",
    "NDArrayDM3",
    "NDArrayDms",
    "NDArrayDms0",
    "NDArrayDms1",
    "NDArrayDms2",
    "NDArrayDms3",
    "NDArrayDns",
    "NDArrayDns0",
    "NDArrayDns1",
    "NDArrayDns2",
    "NDArrayDns3",
    "NDArrayDps",
    "NDArrayDps0",
    "NDArrayDps1",
    "NDArrayDps2",
    "NDArrayDps3",
    "NDArrayDs",
    "NDArrayDs0",
    "NDArrayDs1",
    "NDArrayDs2",
    "NDArrayDs3",
    "NDArrayDus",
    "NDArrayDus0",
    "NDArrayDus1",
    "NDArrayDus2",
    "NDArrayDus3",
    "NDArrayDW",
    "NDArrayDW0",
    "NDArrayDW1",
    "NDArrayDW2",
    "NDArrayDW3",
    "NDArrayDY",
    "NDArrayDY0",
    "NDArrayDY1",
    "NDArrayDY2",
    "NDArrayDY3",
    "NDArrayF",
    "NDArrayF0",
    "NDArrayF0Fin",
    "NDArrayF0FinInt",
    "NDArrayF0FinIntNan",
    "NDArrayF0FinNan",
    "NDArrayF0FinNeg",
    "NDArrayF0FinNegNan",
    "NDArrayF0FinNonNeg",
    "NDArrayF0FinNonNegNan",
    "NDArrayF0FinNonPos",
    "NDArrayF0FinNonPosNan",
    "NDArrayF0FinNonZr",
    "NDArrayF0FinNonZrNan",
    "NDArrayF0FinPos",
    "NDArrayF0FinPosNan",
    "NDArrayF0Int",
    "NDArrayF0IntNan",
    "NDArrayF0Neg",
    "NDArrayF0NegNan",
    "NDArrayF0NonNeg",
    "NDArrayF0NonNegNan",
    "NDArrayF0NonPos",
    "NDArrayF0NonPosNan",
    "NDArrayF0NonZr",
    "NDArrayF0NonZrNan",
    "NDArrayF0Pos",
    "NDArrayF0PosNan",
    "NDArrayF0Zr",
    "NDArrayF0ZrFinNonMic",
    "NDArrayF0ZrFinNonMicNan",
    "NDArrayF0ZrNan",
    "NDArrayF0ZrNonMic",
    "NDArrayF0ZrNonMicNan",
    "NDArrayF1",
    "NDArrayF1Fin",
    "NDArrayF1FinInt",
    "NDArrayF1FinIntNan",
    "NDArrayF1FinNan",
    "NDArrayF1FinNeg",
    "NDArrayF1FinNegNan",
    "NDArrayF1FinNonNeg",
    "NDArrayF1FinNonNegNan",
    "NDArrayF1FinNonPos",
    "NDArrayF1FinNonPosNan",
    "NDArrayF1FinNonZr",
    "NDArrayF1FinNonZrNan",
    "NDArrayF1FinPos",
    "NDArrayF1FinPosNan",
    "NDArrayF1Int",
    "NDArrayF1IntNan",
    "NDArrayF1Neg",
    "NDArrayF1NegNan",
    "NDArrayF1NonNeg",
    "NDArrayF1NonNegNan",
    "NDArrayF1NonPos",
    "NDArrayF1NonPosNan",
    "NDArrayF1NonZr",
    "NDArrayF1NonZrNan",
    "NDArrayF1Pos",
    "NDArrayF1PosNan",
    "NDArrayF1Zr",
    "NDArrayF1ZrFinNonMic",
    "NDArrayF1ZrFinNonMicNan",
    "NDArrayF1ZrNan",
    "NDArrayF1ZrNonMic",
    "NDArrayF1ZrNonMicNan",
    "NDArrayF2",
    "NDArrayF2Fin",
    "NDArrayF2FinInt",
    "NDArrayF2FinIntNan",
    "NDArrayF2FinNan",
    "NDArrayF2FinNeg",
    "NDArrayF2FinNegNan",
    "NDArrayF2FinNonNeg",
    "NDArrayF2FinNonNegNan",
    "NDArrayF2FinNonPos",
    "NDArrayF2FinNonPosNan",
    "NDArrayF2FinNonZr",
    "NDArrayF2FinNonZrNan",
    "NDArrayF2FinPos",
    "NDArrayF2FinPosNan",
    "NDArrayF2Int",
    "NDArrayF2IntNan",
    "NDArrayF2Neg",
    "NDArrayF2NegNan",
    "NDArrayF2NonNeg",
    "NDArrayF2NonNegNan",
    "NDArrayF2NonPos",
    "NDArrayF2NonPosNan",
    "NDArrayF2NonZr",
    "NDArrayF2NonZrNan",
    "NDArrayF2Pos",
    "NDArrayF2PosNan",
    "NDArrayF2Zr",
    "NDArrayF2ZrFinNonMic",
    "NDArrayF2ZrFinNonMicNan",
    "NDArrayF2ZrNan",
    "NDArrayF2ZrNonMic",
    "NDArrayF2ZrNonMicNan",
    "NDArrayF3",
    "NDArrayF3Fin",
    "NDArrayF3FinInt",
    "NDArrayF3FinIntNan",
    "NDArrayF3FinNan",
    "NDArrayF3FinNeg",
    "NDArrayF3FinNegNan",
    "NDArrayF3FinNonNeg",
    "NDArrayF3FinNonNegNan",
    "NDArrayF3FinNonPos",
    "NDArrayF3FinNonPosNan",
    "NDArrayF3FinNonZr",
    "NDArrayF3FinNonZrNan",
    "NDArrayF3FinPos",
    "NDArrayF3FinPosNan",
    "NDArrayF3Int",
    "NDArrayF3IntNan",
    "NDArrayF3Neg",
    "NDArrayF3NegNan",
    "NDArrayF3NonNeg",
    "NDArrayF3NonNegNan",
    "NDArrayF3NonPos",
    "NDArrayF3NonPosNan",
    "NDArrayF3NonZr",
    "NDArrayF3NonZrNan",
    "NDArrayF3Pos",
    "NDArrayF3PosNan",
    "NDArrayF3Zr",
    "NDArrayF3ZrFinNonMic",
    "NDArrayF3ZrFinNonMicNan",
    "NDArrayF3ZrNan",
    "NDArrayF3ZrNonMic",
    "NDArrayF3ZrNonMicNan",
    "NDArrayFFin",
    "NDArrayFFinInt",
    "NDArrayFFinIntNan",
    "NDArrayFFinNan",
    "NDArrayFFinNeg",
    "NDArrayFFinNegNan",
    "NDArrayFFinNonNeg",
    "NDArrayFFinNonNegNan",
    "NDArrayFFinNonPos",
    "NDArrayFFinNonPosNan",
    "NDArrayFFinNonZr",
    "NDArrayFFinNonZrNan",
    "NDArrayFFinPos",
    "NDArrayFFinPosNan",
    "NDArrayFInt",
    "NDArrayFIntNan",
    "NDArrayFNeg",
    "NDArrayFNegNan",
    "NDArrayFNonNeg",
    "NDArrayFNonNegNan",
    "NDArrayFNonPos",
    "NDArrayFNonPosNan",
    "NDArrayFNonZr",
    "NDArrayFNonZrNan",
    "NDArrayFPos",
    "NDArrayFPosNan",
    "NDArrayFZr",
    "NDArrayFZrFinNonMic",
    "NDArrayFZrFinNonMicNan",
    "NDArrayFZrNan",
    "NDArrayFZrNonMic",
    "NDArrayFZrNonMicNan",
    "NDArrayI",
    "NDArrayI0",
    "NDArrayI0Neg",
    "NDArrayI0NonNeg",
    "NDArrayI0NonPos",
    "NDArrayI0NonZr",
    "NDArrayI0Pos",
    "NDArrayI0Zr",
    "NDArrayI1",
    "NDArrayI1Neg",
    "NDArrayI1NonNeg",
    "NDArrayI1NonPos",
    "NDArrayI1NonZr",
    "NDArrayI1Pos",
    "NDArrayI1Zr",
    "NDArrayI2",
    "NDArrayI2Neg",
    "NDArrayI2NonNeg",
    "NDArrayI2NonPos",
    "NDArrayI2NonZr",
    "NDArrayI2Pos",
    "NDArrayI2Zr",
    "NDArrayI3",
    "NDArrayI3Neg",
    "NDArrayI3NonNeg",
    "NDArrayI3NonPos",
    "NDArrayI3NonZr",
    "NDArrayI3Pos",
    "NDArrayI3Zr",
    "NDArrayINeg",
    "NDArrayINonNeg",
    "NDArrayINonPos",
    "NDArrayINonZr",
    "NDArrayIPos",
    "NDArrayIZr",
    "NDArrayO",
    "NDArrayO0",
    "NDArrayO1",
    "NDArrayO2",
    "NDArrayO3",
    "NDim0",
    "NDim1",
    "NDim2",
    "NDim3",
    "NonIntegralElementsError",
    "NoTrueElementsError",
    "redirect_to_empty_numpy_concatenate_error",
    "shift_bool",
    "shift",
    "year",
    "ZeroShiftError",
]

try:
    from utilities.numpy.bottleneck import (
        ZeroPercentageChangeSpanError,
        ffill,
        pct_change,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["ffill", "pct_change", "ZeroPercentageChangeSpanError"]


try:
    from utilities.numpy.numbagg import ewma, exp_moving_sum
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:
    __all__ += ["ewma", "exp_moving_sum"]



from __future__ import annotations

import abc
from abc import ABC, ABCMeta
from ast import ImportFrom, alias

import pytest
from click.testing import CliRunner
from pytest import LogCaptureFixture, mark, param

from utilities.itertools import one
from utilities.scripts.generate_snippets import (
    Method,
    _generate_ipython_imports,
    _generate_snippet,
    _generate_snippets,
    _node_to_key,
    _yield_import_nodes_directly,
    _yield_import_nodes_from_module_all,
    _yield_import_nodes_from_text,
    main,
    yield_imports,
)


class TestCLI:
    def test_main(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main)
        assert result.exit_code == 0


class TestGenerateIPythonImports:
    def test_main(self) -> None:
        imp = ImportFrom(module="abc", names=[alias(name="ABC")])
        result = _generate_ipython_imports([imp])
        expected = "from abc import ABC  # noqa: F401"
        assert result == expected


class TestGenerateSnippet:
    def test_main(self) -> None:
        imp = ImportFrom(module="abc", names=[alias(name="ABC")])
        template = "{key}: {value}"
        result = _generate_snippet(imp, template)
        expected = "fab-abc: from abc import ABC"
        assert result == expected

    def test_as(self) -> None:
        imp = ImportFrom(
            module="collections.abc", names=[alias(name="Set", asname="AbstractSet")]
        )
        template = "{key}: {value}"
        result = _generate_snippet(imp, template)
        expected = "fco-set: from collections.abc import Set as AbstractSet"
        assert result == expected


class TestGenerateSnippets:
    def test_main(self) -> None:
        imports = {
            ImportFrom(module="abc", names=[alias(name=name)])
            for name in ["ABC", "ABCMeta"]
        }
        template = "{key}: {value}"
        result = _generate_snippets(imports, template)
        expected = "fab-abc: from abc import ABC,fab-abc-meta: from abc import ABCMeta,"
        assert result == expected

    def test_duplicated_keys(self, *, caplog: LogCaptureFixture) -> None:
        imports = {
            ImportFrom(module="dataclasses", names=[alias(name=name)])
            for name in ["field", "Field"]
        }
        template = "{key}: {value}"
        _ = _generate_snippets(imports, template)
        (record,) = caplog.records
        assert record.message == "Duplicated keys: {'fda-field'}"


class TestNodeToKey:
    @mark.parametrize(
        ("module", "name", "expected"),
        [param("abc", "ABC", "fab-abc"), param("abc", "ABCMeta", "fab-abc-meta")],
    )
    def test_main(self, *, module: str, name: str, expected: str) -> None:
        node = ImportFrom(module=module, names=[alias(name=name)])
        key = _node_to_key(node)
        assert key == expected


class TestYieldImportNodesDirectly:
    def test_main(self) -> None:
        imp1, imp2 = _yield_import_nodes_directly(abc, [ABC, ABCMeta])
        assert imp1.module == "abc"
        assert one(imp1.names).name == "ABC"
        assert imp2.module == "abc"
        assert one(imp2.names).name == "ABCMeta"


class TestYieldImportNodesFromModuleAll:
    def test_main(self) -> None:
        imports = list(_yield_import_nodes_from_module_all(pytest))
        assert len(imports) == 77
        for imp in imports:
            assert imp.module == "pytest"


class TestYieldImportNodesFromText:
    def test_single(self) -> None:
        text = """
            from abc import ABC
        """
        imp = one(_yield_import_nodes_from_text(text))
        assert imp.module == "abc"
        assert one(imp.names).name == "ABC"

    @mark.parametrize(
        "text",
        [
            param(
                """
                from abc import ABC
                from abc import ABCMeta
                """
            ),
            param(
                """
                from abc import (
                    ABC,
                    ABCMeta,
                )
                """
            ),
        ],
    )
    def test_two(self, *, text: str) -> None:
        imp1, imp2 = _yield_import_nodes_from_text(text)
        assert imp1.module == "abc"
        assert one(imp1.names).name == "ABC"
        assert imp2.module == "abc"
        assert one(imp2.names).name == "ABCMeta"

    def test_suppress_include(self) -> None:
        text = """
            from contextlib import suppress

            with suppress(ModuleNotFoundError):
                from abc import ABC
        """
        (imp1, imp2) = _yield_import_nodes_from_text(text, include_suppress=True)
        assert imp1.module == "contextlib"
        assert one(imp1.names).name == "suppress"
        assert imp2.module == "abc"
        assert one(imp2.names).name == "ABC"

    def test_suppress_do_not_include(self) -> None:
        text = """
            from contextlib import suppress

            with suppress(ModuleNotFoundError):
                from abc import ABC
        """
        (imp,) = _yield_import_nodes_from_text(text, include_suppress=False)
        assert imp.module == "abc"
        assert one(imp.names).name == "ABC"

    @mark.parametrize(
        "text",
        [
            param(
                """
                from contextlib import suppress

                with suppress(ModuleNotFoundError):
                    from abc import ABC
                    with suppress(ModuleNotFoundError):
                        from abc import ABCMeta
                """
            ),
            param(
                """
                from contextlib import suppress

                with suppress(ModuleNotFoundError):
                    with suppress(ModuleNotFoundError):
                        from abc import ABC
                    from abc import ABCMeta
                """
            ),
        ],
    )
    def test_nested_suppress(self, *, text: str) -> None:
        imp1, imp2 = _yield_import_nodes_from_text(text)
        assert imp1.module == "abc"
        assert one(imp1.names).name == "ABC"
        assert imp2.module == "abc"
        assert one(imp2.names).name == "ABCMeta"


class TestYieldImports:
    @mark.parametrize("method", [param(Method.direct), param(Method.parse)])
    def test_yield_imports(self, *, method: Method) -> None:
        for imp in yield_imports(method=method):
            assert isinstance(imp, ImportFrom)



from __future__ import annotations

from sqlalchemy import Engine

from utilities.luigi import EngineParameter
from utilities.luigi.typed_settings import _map_annotation


class TestMapAnnotation:
    def test_main(self) -> None:
        result = _map_annotation(Engine)
        param = result()
        assert isinstance(param, EngineParameter)



from __future__ import annotations

from pydantic import BaseModel
from typing_extensions import override


class HashableBaseModel(BaseModel):
    """Subclass of BaseModel which is hashable."""

    @override
    def __hash__(self) -> int:
        return hash((type(self), *self.__dict__.values()))



from pathlib import Path

from hypothesis import given
from sqlalchemy import Engine

from tests.typed_settings.test_click import TestClickOptions as _TestClickOptions
from tests.typed_settings.test_typed_settings import (
    TestLoadSettings as _TestLoadSettings,
)
from utilities.hypothesis import sqlite_engines, temp_paths, text_ascii
from utilities.pytest import skipif_windows
from utilities.sqlalchemy import serialize_engine

app_names = text_ascii(min_size=1).map(str.lower)


class TestLoadSettings:
    @skipif_windows  # writing \\ to file
    @given(
        default=sqlite_engines(),
        appname=app_names,
        root=temp_paths(),
        value=sqlite_engines(),
    )
    def test_main(
        self, *, default: Engine, root: Path, appname: str, value: Engine
    ) -> None:
        def equal(x: Engine, y: Engine, /) -> bool:
            return x.url == y.url

        _TestLoadSettings.run_test(
            Engine, default, root, appname, serialize_engine, value, equal
        )


class TestClickOptions:
    @skipif_windows  # writing \\ to file
    @given(
        default=sqlite_engines(),
        appname=app_names,
        root=temp_paths(),
        value=sqlite_engines(),
        cfg=sqlite_engines(),
    )
    def test_main(
        self, *, default: Engine, appname: str, root: Path, value: Engine, cfg: Engine
    ) -> None:
        _TestClickOptions.run_test(
            Engine, default, appname, serialize_engine, root, value, cfg
        )



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import integers

from utilities.pydantic import HashableBaseModel


class TestHashableBaseModel:
    @given(x=integers())
    def test_main(self, *, x: int) -> None:
        class Example(HashableBaseModel):
            x: int

        example = Example(x=x)
        assert isinstance(hash(example), int)



from __future__ import annotations

from hypothesis import assume
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import DrawFn, composite, integers
from semver import Version

from utilities.hypothesis.hypothesis import (
    MaybeSearchStrategy,
    lift_draw,
    lists_fixed_length,
)


@composite
def versions(  # noqa: PLR0912
    _draw: DrawFn,
    /,
    *,
    min_version: MaybeSearchStrategy[Version | None] = None,
    max_version: MaybeSearchStrategy[Version | None] = None,
) -> Version:
    """Strategy for generating `Version`s."""
    draw = lift_draw(_draw)
    min_version_, max_version_ = (draw(mv) for mv in (min_version, max_version))
    if isinstance(min_version_, Version) and isinstance(max_version_, Version):
        if min_version_ > max_version_:
            msg = f"{min_version_=}, {max_version_=}"
            raise InvalidArgument(msg)
        major = draw(integers(min_version_.major, max_version_.major))
        minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        version = Version(major=major, minor=minor, patch=patch)
        _ = assume(min_version_ <= version <= max_version_)
        return version
    if isinstance(min_version_, Version) and (max_version_ is None):
        major = draw(integers(min_value=min_version_.major))
        if major > min_version_.major:
            minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        else:
            minor = draw(integers(min_version_.minor))
            if minor > min_version_.minor:
                patch = draw(integers(min_value=0))  # pragma: no cover
            else:
                patch = draw(integers(min_value=min_version_.patch))
    elif (min_version_ is None) and isinstance(max_version_, Version):
        major = draw(integers(0, max_version_.major))
        if major < max_version_.major:
            minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        else:
            minor = draw(integers(0, max_version_.minor))
            if minor < max_version_.minor:
                patch = draw(integers(min_value=0))  # pragma: no cover
            else:
                patch = draw(integers(0, max_version_.patch))
    elif (min_version_ is None) and (max_version_ is None):
        major, minor, patch = draw(lists_fixed_length(integers(min_value=0), 3))
    else:
        msg = "Invalid case"  # pragma: no cover
        raise RuntimeError(msg)  # pragma: no cover
    return Version(major=major, minor=minor, patch=patch)


__all__ = ["versions"]



from __future__ import annotations

from utilities.click.click import (
    Date,
    DateTime,
    Enum,
    Time,
    Timedelta,
    log_level_option,
)

__all__ = ["Date", "DateTime", "Enum", "log_level_option", "Time", "Timedelta"]


try:
    from utilities.click.luigi import (
        local_scheduler_option_default_central,
        local_scheduler_option_default_local,
        workers_option,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "local_scheduler_option_default_central",
        "local_scheduler_option_default_local",
        "workers_option",
    ]


try:
    from utilities.click.sqlalchemy import Engine
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["Engine"]



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from itertools import chain
from typing import Any

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    binary,
    data,
    dictionaries,
    integers,
    lists,
    sampled_from,
    sets,
    text,
)
from pytest import mark, param, raises

from utilities.itertools import (
    EmptyIterableError,
    IterableContainsDuplicatesError,
    MultipleElementsError,
    always_iterable,
    check_duplicates,
    chunked,
    is_iterable_not_str,
    one,
    take,
)


class TestAlwaysIterable:
    @given(x=binary())
    def test_bytes(self, *, x: bytes) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == list(x)
        assert list(always_iterable(x, base_type=bytes)) == [x]
        assert list(always_iterable(x, base_type=(bytes,))) == [x]

    @given(x=integers())
    def test_integer(self, *, x: int) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == [x]
        assert list(always_iterable(x, base_type=int)) == [x]
        assert list(always_iterable(x, base_type=(int,))) == [x]

    @given(x=text())
    def test_string(self, *, x: str) -> None:
        assert list(always_iterable(x)) == [x]
        assert list(always_iterable(x, base_type=None)) == list(x)
        assert list(always_iterable(x, base_type=str)) == [x]
        assert list(always_iterable(x, base_type=(str,))) == [x]

    @given(x=dictionaries(text(), integers()))
    def test_dict(self, *, x: dict[str, int]) -> None:
        assert list(always_iterable(x)) == list(x)
        assert list(always_iterable(x, base_type=dict)) == [x]
        assert list(always_iterable(x, base_type=(dict,))) == [x]

    @given(x=lists(integers()))
    def test_lists(self, *, x: list[int]) -> None:
        assert list(always_iterable(x)) == x
        assert list(always_iterable(x, base_type=None)) == x
        assert list(always_iterable(x, base_type=list)) == [x]
        assert list(always_iterable(x, base_type=(list,))) == [x]

    def test_none(self) -> None:
        assert list(always_iterable(None)) == []

    def test_generator(self) -> None:
        def yield_ints() -> Iterator[int]:
            yield 0
            yield 1

        assert list(always_iterable(yield_ints())) == [0, 1]


class TestCheckDuplicates:
    @given(x=sets(integers()))
    def test_main(self, *, x: set[int]) -> None:
        check_duplicates(x)

    @given(data=data(), x=lists(integers(), min_size=1))
    def test_error(self, *, data: DataObject, x: Sequence[int]) -> None:
        x_i = data.draw(sampled_from(x))
        y = chain(x, [x_i])
        with raises(IterableContainsDuplicatesError):
            check_duplicates(y)


class TestChunked:
    @mark.parametrize(
        ("iterable", "expected"),
        [
            param([1, 2, 3, 4, 5, 6], [[1, 2, 3], [4, 5, 6]]),
            param([1, 2, 3, 4, 5, 6, 7, 8], [[1, 2, 3], [4, 5, 6], [7, 8]]),
        ],
    )
    def test_main(self, *, iterable: list[int], expected: list[list[int]]) -> None:
        result = list(chunked(iterable, n=3))
        assert result == expected


class TestIsIterableNotStr:
    @mark.parametrize(
        ("x", "expected"),
        [param(None, False), param([], True), param((), True), param("", False)],
    )
    def test_main(self, *, x: Any, expected: bool) -> None:
        assert is_iterable_not_str(x) is expected


class TestOne:
    def test_empty(self) -> None:
        with raises(EmptyIterableError):
            _ = one([])

    def test_one(self) -> None:
        assert one([None]) is None

    def test_multiple(self) -> None:
        with raises(
            MultipleElementsError,
            match="Expected exactly one item in iterable, but got 1, 2, and "
            "perhaps more",
        ):
            _ = one([1, 2, 3])


class TestTake:
    @mark.parametrize(("n", "iterable"), [param(3, range(10)), param(10, range(3))])
    def test_main(self, *, n: int, iterable: Iterable[int]) -> None:
        result = take(n, iterable)
        assert result == [0, 1, 2]



from __future__ import annotations

import datetime as dt
from typing import Any, cast

import numpy as np
from hypothesis import assume
from hypothesis.errors import InvalidArgument
from hypothesis.extra.numpy import array_shapes, arrays
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    booleans,
    composite,
    integers,
    none,
    nothing,
    sampled_from,
)
from numpy import (
    concatenate,
    datetime64,
    expand_dims,
    iinfo,
    int32,
    int64,
    uint32,
    uint64,
    zeros,
)
from numpy.typing import NDArray

from utilities.hypothesis.hypothesis import (
    MaybeSearchStrategy,
    Shape,
    floats_extra,
    lift_draw,
    lists_fixed_length,
    text_ascii,
)
from utilities.math import IntNonNeg
from utilities.numpy import (
    DATE_MAX_AS_INT,
    DATE_MIN_AS_INT,
    DATETIME_MAX_AS_INT,
    DATETIME_MIN_AS_INT,
    Datetime64Kind,
    Datetime64Unit,
    EmptyNumpyConcatenateError,
    NDArrayB,
    NDArrayD,
    NDArrayD1,
    NDArrayDD1,
    NDArrayDus1,
    NDArrayF,
    NDArrayI,
    NDArrayO,
    date_to_datetime64,
    datetime64_to_int,
    datetime64_unit_to_dtype,
    datetime64_unit_to_kind,
    datetime_to_datetime64,
    redirect_to_empty_numpy_concatenate_error,
)


@composite
def bool_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayB:
    """Strategy for generating arrays of booleans."""
    draw = lift_draw(_draw)
    strategy = cast(
        SearchStrategy[NDArrayB],
        arrays(bool, draw(shape), elements=booleans(), fill=fill, unique=draw(unique)),
    )
    return draw(strategy)


@composite
def concatenated_arrays(
    _draw: DrawFn,
    strategy: SearchStrategy[NDArray[Any]],
    size: MaybeSearchStrategy[IntNonNeg],
    fallback: Shape,
    /,
    *,
    dtype: Any = float,
) -> NDArray[Any]:
    """Strategy for generating arrays from lower-dimensional strategies."""
    draw = lift_draw(_draw)
    size_ = draw(size)
    arrays = draw(lists_fixed_length(strategy, size_))
    expanded = [expand_dims(array, axis=0) for array in arrays]
    try:
        return concatenate(expanded)
    except ValueError as error:
        try:
            redirect_to_empty_numpy_concatenate_error(error)
        except EmptyNumpyConcatenateError:
            if isinstance(fallback, int):
                shape = size_, fallback
            else:
                shape = (size_, *fallback)
            return zeros(shape, dtype=dtype)


@composite
def datetime64_dtypes(
    _draw: DrawFn, /, *, kind: MaybeSearchStrategy[Datetime64Kind | None] = None
) -> Any:
    """Strategy for generating datetime64 dtypes."""
    draw = lift_draw(_draw)
    unit = draw(datetime64_units(kind=kind))
    return datetime64_unit_to_dtype(unit)


def datetime64_kinds() -> SearchStrategy[Datetime64Kind]:
    """Strategy for generating datetime64 kinds."""
    kinds: list[Datetime64Kind] = ["date", "time"]
    return sampled_from(kinds)


@composite
def datetime64_units(
    _draw: DrawFn, /, *, kind: MaybeSearchStrategy[Datetime64Kind | None] = None
) -> Datetime64Unit:
    """Strategy for generating datetime64 units."""
    draw = lift_draw(_draw)
    units: list[Datetime64Unit] = [
        "Y",
        "M",
        "W",
        "D",
        "h",
        "m",
        "s",
        "ms",
        "us",
        "ns",
        "ps",
        "fs",
        "as",
    ]
    kind_ = draw(kind)
    if kind_ is not None:
        units = [unit for unit in units if datetime64_unit_to_kind(unit) == kind_]
    return draw(sampled_from(units))


@composite
def datetime64_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayD:
    """Strategy for generating arrays of datetime64s."""
    draw = lift_draw(_draw)
    if (unit_ := draw(unit)) is None:
        unit_ = draw(datetime64_units())
    dtype = datetime64_unit_to_dtype(cast(Datetime64Unit, unit_))
    elements = datetime64s(
        unit=cast(Datetime64Unit, unit_),
        min_value=min_value,
        max_value=max_value,
        valid_dates=valid_dates,
        valid_datetimes=valid_datetimes,
    )
    return draw(
        arrays(dtype, draw(shape), elements=elements, fill=fill, unique=draw(unique))
    )


@composite
def datetime64_indexes(
    _draw: DrawFn,
    /,
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> NDArrayD1:
    """Strategy for generating indexes of datetime64s."""
    draw = lift_draw(_draw)
    array = draw(
        datetime64_arrays(
            shape=n,
            unit=unit,
            min_value=min_value,
            max_value=max_value,
            valid_dates=valid_dates,
            valid_datetimes=valid_datetimes,
            fill=nothing(),
            unique=unique,
        )
    )
    return np.sort(array) if sort else array


def datetime64D_indexes(  # noqa: N802
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = True,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> SearchStrategy[NDArrayDD1]:
    """Strategy for generating arrays of dates."""
    return datetime64_indexes(
        n=n,
        unit="D",
        min_value=min_value,
        max_value=max_value,
        valid_dates=valid_dates,
        unique=unique,
        sort=sort,
    )


@composite
def datetime64s(
    _draw: DrawFn,
    /,
    *,
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
) -> datetime64:
    """Strategy for generating datetime64s."""
    draw = lift_draw(_draw)
    unit_ = draw(unit)
    min_value_, max_value_ = (
        _datetime64s_convert(draw(mv)) for mv in (min_value, max_value)
    )
    if draw(valid_dates):
        unit_, min_value_, max_value_ = _datetime64s_check_valid_dates(
            unit=cast(Datetime64Unit | None, unit_),
            min_value=min_value_,
            max_value=max_value_,
        )
    if draw(valid_datetimes):
        unit_, min_value_, max_value_ = _datetime64s_check_valid_datetimes(
            unit=cast(Datetime64Unit | None, unit_),
            min_value=min_value_,
            max_value=max_value_,
        )
    i = draw(int64s(min_value=min_value_, max_value=max_value_))
    _ = assume(i != iinfo(int64).min)
    if unit_ is None:
        unit_ = draw(datetime64_units())
    return datetime64(i, unit_)


def _datetime64s_convert(value: int | datetime64 | dt.date | None, /) -> int | None:
    """Convert a min/max value supplied into `datetime64s`."""
    if (value is None) or isinstance(value, int):
        return value
    if isinstance(value, datetime64):
        return datetime64_to_int(value)
    if isinstance(value, dt.datetime):
        return _datetime64s_convert(datetime_to_datetime64(value))
    return _datetime64s_convert(date_to_datetime64(value))


def _datetime64s_check_valid_dates(
    *,
    unit: Datetime64Unit | None = None,
    min_value: int | None = None,
    max_value: int | None = None,
) -> tuple[Datetime64Unit, int | None, int | None]:
    """Check/clip the bounds to generate valid `dt.date`s."""
    if (unit is not None) and (unit != "D"):
        msg = f"{unit=}"
        raise InvalidArgument(msg)
    if min_value is None:
        min_value = DATE_MIN_AS_INT
    else:
        min_value = max(min_value, DATE_MIN_AS_INT)
    if max_value is None:
        max_value = DATE_MAX_AS_INT
    else:
        max_value = min(max_value, DATE_MAX_AS_INT)
    return "D", min_value, max_value


def _datetime64s_check_valid_datetimes(
    *,
    unit: Datetime64Unit | None = None,
    min_value: int | None = None,
    max_value: int | None = None,
) -> tuple[Datetime64Unit, int | None, int | None]:
    """Check/clip the bounds to generate valid `dt.datetime`s."""
    if (unit is not None) and (unit != "us"):
        msg = f"{unit=}"
        raise InvalidArgument(msg)
    if min_value is None:
        min_value = DATETIME_MIN_AS_INT
    else:
        min_value = max(min_value, DATETIME_MIN_AS_INT)
    if max_value is None:
        max_value = DATETIME_MAX_AS_INT
    else:
        max_value = min(max_value, DATETIME_MAX_AS_INT)
    return "us", min_value, max_value


def datetime64us_indexes(
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    min_value: MaybeSearchStrategy[datetime64 | int | dt.datetime | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.datetime | None] = None,
    valid_datetimes: MaybeSearchStrategy[bool] = True,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> SearchStrategy[NDArrayDus1]:
    """Strategy for generating arrays of datetimes."""
    return datetime64_indexes(
        n=n,
        unit="us",
        min_value=min_value,
        max_value=max_value,
        valid_datetimes=valid_datetimes,
        unique=unique,
        sort=sort,
    )


@composite
def float_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayF:
    """Strategy for generating arrays of floats."""
    draw = lift_draw(_draw)
    elements = floats_extra(
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_inf=allow_inf,
        allow_pos_inf=allow_pos_inf,
        allow_neg_inf=allow_neg_inf,
        integral=integral,
    )
    strategy = cast(
        SearchStrategy[NDArrayF],
        arrays(float, draw(shape), elements=elements, fill=fill, unique=draw(unique)),
    )
    return draw(strategy)


@composite
def int_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayI:
    """Strategy for generating arrays of ints."""
    draw = lift_draw(_draw)
    info = iinfo(int64)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_use = info.min if min_value_ is None else min_value_
    max_value_use = info.max if max_value_ is None else max_value_
    elements = integers(min_value=min_value_use, max_value=max_value_use)
    strategy = cast(
        SearchStrategy[NDArrayI],
        arrays(int64, draw(shape), elements=elements, fill=fill, unique=draw(unique)),
    )
    return draw(strategy)


def int32s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating int32s."""
    return _fixed_width_ints(int32, min_value=min_value, max_value=max_value)


def int64s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating int64s."""
    return _fixed_width_ints(int64, min_value=min_value, max_value=max_value)


@composite
def str_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    allow_none: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayO:
    """Strategy for generating arrays of strings."""
    draw = lift_draw(_draw)
    elements = text_ascii(min_size=min_size, max_size=max_size)
    if draw(allow_none):
        elements |= none()
    strategy = cast(
        SearchStrategy[NDArrayO],
        arrays(object, draw(shape), elements=elements, fill=fill, unique=draw(unique)),
    )
    return draw(strategy)


def uint32s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating uint32s."""
    return _fixed_width_ints(uint32, min_value=min_value, max_value=max_value)


def uint64s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating uint64s."""
    return _fixed_width_ints(uint64, min_value=min_value, max_value=max_value)


@composite
def _fixed_width_ints(
    _draw: DrawFn,
    dtype: Any,
    /,
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> int:
    """Strategy for generating int64s."""
    draw = lift_draw(_draw)
    min_value_, max_value_ = (draw(mv) for mv in (min_value, max_value))
    info = iinfo(dtype)
    min_value_ = info.min if min_value_ is None else max(info.min, min_value_)
    max_value = info.max if max_value_ is None else min(info.max, max_value_)
    return draw(integers(min_value_, max_value))


__all__ = [
    "bool_arrays",
    "concatenated_arrays",
    "datetime64_arrays",
    "datetime64_dtypes",
    "datetime64_indexes",
    "datetime64_kinds",
    "datetime64_units",
    "datetime64D_indexes",
    "datetime64s",
    "datetime64us_indexes",
    "float_arrays",
    "int_arrays",
    "int32s",
    "int64s",
    "str_arrays",
    "uint32s",
    "uint64s",
]



from __future__ import annotations

from typing import Any

from typing_extensions import override


class _Meta(type):
    """Metaclass for the sentinel."""

    instance: Any = None

    @override
    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


_REPR = "<sentinel>"


class Sentinel(metaclass=_Meta):
    """Base class for the sentinel object."""

    @override
    def __repr__(self) -> str:
        return _REPR

    @override
    def __str__(self) -> str:
        return repr(self)


sentinel = Sentinel()


__all__ = ["sentinel", "Sentinel"]



from __future__ import annotations

from utilities.typed_settings.typed_settings import (
    AppNameContainsUnderscoreError,
    get_repo_root_config,
    load_settings,
)

__all__ = ["AppNameContainsUnderscoreError", "get_repo_root_config", "load_settings"]


try:
    from utilities.typed_settings.click import click_field, click_options
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["click_field", "click_options"]



from __future__ import annotations

import datetime as dt
from collections.abc import Iterator
from datetime import timezone
from math import nan
from typing import Any, Literal, cast, overload

import polars as pl
from polars import (
    DataFrame,
    Date,
    Datetime,
    Expr,
    Float64,
    Int64,
    PolarsDataType,
    Utf8,
    col,
    read_database,
    when,
)
from polars.type_aliases import ConnectionOrCursor, SchemaDict
from sqlalchemy import Column, Connection, Engine, Select, Table
from sqlalchemy.sql.base import ReadOnlyColumnCollection

from utilities.datetime import UTC
from utilities.sqlalchemy import (
    CHUNK_SIZE_FRAC,
    check_dataframe_schema_against_table,
    check_selectable_for_duplicate_columns,
    insert_items,
    yield_connection,
)
from utilities.text import snake_case


def insert_polars_dataframe(
    engine_or_conn: Engine | Connection,
    df: DataFrame,
    table_or_mapped_class: Table | type[Any],
    /,
    *,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> None:
    """Insert a DataFrame into a database."""
    mapping = check_dataframe_schema_against_table(
        dict(df.schema), table_or_mapped_class, _check_polars_series, snake=snake
    )
    items = df.select(mapping).rename(mapping).to_dicts()
    if (df.height > 0) and (len(items) == 0):
        msg = f"{df=}, {items=}"
        raise PolarsDataFrameYieldsNoRowsError(msg)
    return insert_items(
        engine_or_conn, (items, table_or_mapped_class), chunk_size_frac=chunk_size_frac
    )


class PolarsDataFrameYieldsNoRowsError(ValueError):
    """Raised when a DataFrame yields no rows to insert."""


def _check_polars_series(dtype: PolarsDataType, py_type: type, /) -> bool:
    return (
        (dtype is pl.Boolean and issubclass(py_type, bool))
        or (
            dtype is Date
            and issubclass(py_type, dt.date)
            and not issubclass(py_type, dt.datetime)
        )
        or (isinstance(dtype, Datetime) and issubclass(py_type, dt.datetime))
        or (dtype is Float64 and issubclass(py_type, float))
        or (dtype is Int64 and issubclass(py_type, int))
        or (dtype is Utf8 and issubclass(py_type, str))
    )


@overload
def select_to_polars_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    iter_batches: Literal[True],
    batch_size: int | None = ...,
    schema_overrides: SchemaDict | None = ...,
    time_zone: timezone = ...,
    snake: bool = ...,
    **kwargs: Any,
) -> Iterator[DataFrame]:
    ...


@overload
def select_to_polars_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    iter_batches: Literal[False] = ...,
    batch_size: int | None = ...,
    schema_overrides: SchemaDict | None = ...,
    time_zone: timezone = ...,
    snake: bool = ...,
    **kwargs: Any,
) -> DataFrame:
    ...


def select_to_polars_dataframe(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    iter_batches: bool = False,
    batch_size: int | None = None,
    schema_overrides: SchemaDict | None = None,
    time_zone: timezone = UTC,
    snake: bool = False,
    **kwargs: Any,
) -> DataFrame | Iterator[DataFrame]:
    """Read a table from a database into a DataFrame.

    Optionally stream it in chunks.
    """
    check_selectable_for_duplicate_columns(sel)
    if iter_batches:
        return _stream_dataframes(
            sel,
            engine_or_conn,
            batch_size=batch_size,
            schema_overrides=schema_overrides,
            time_zone=time_zone,
            snake=snake,
            **kwargs,
        )
    with yield_connection(engine_or_conn) as conn:
        df = read_database(
            sel,
            cast(ConnectionOrCursor, conn),
            iter_batches=False,
            schema_overrides=schema_overrides,
            **kwargs,
        )
        return _post_process(df, sel, time_zone=time_zone, snake=snake)


def _post_process(
    df: DataFrame,
    sel: Select[Any],
    /,
    *,
    time_zone: timezone = UTC,
    snake: bool = False,
) -> DataFrame:
    """Post-processing for a DataFrame from the database."""
    columns: ReadOnlyColumnCollection = cast(Any, sel).selected_columns
    exprs = (
        _table_column_to_expr(col, time_zone=time_zone).alias(col.name)
        for col in columns
    )
    df = df.with_columns(exprs)
    return _snake_columns(df) if snake else df


def _table_column_to_expr(column: Column[Any], /, *, time_zone: timezone = UTC) -> Expr:
    """Map a table column to a Expr."""
    expr = col(column.name)
    db_type = column.type
    py_type = db_type.python_type
    if issubclass(py_type, bool):
        return expr.cast(pl.Boolean)
    if issubclass(py_type, dt.date) and not issubclass(py_type, dt.datetime):
        return expr.cast(pl.Date)
    if issubclass(py_type, dt.datetime):
        has_tz: bool = cast(Any, db_type).timezone
        sr_dtype = Datetime(time_zone=time_zone) if has_tz else Datetime
        return expr.cast(sr_dtype)
    if issubclass(py_type, float):
        expr_float = expr.cast(Float64)
        return when(expr_float.is_null()).then(nan).otherwise(expr_float)
    if issubclass(py_type, int):
        return expr.cast(Int64)
    if issubclass(py_type, str):
        return expr.cast(Utf8)
    msg = f"{column=}"
    raise ColumnToPolarsExprError(msg)


class ColumnToPolarsExprError(TypeError):
    """Raised when an invalid column type is encountered."""


def _snake_columns(df: DataFrame, /) -> DataFrame:
    mapping = {col: snake_case(col) for col in df.columns}
    return df.rename(mapping)


def _stream_dataframes(
    sel: Select[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    batch_size: int | None = None,
    schema_overrides: SchemaDict | None = None,
    time_zone: timezone = UTC,
    snake: bool = False,
    **kwargs: Any,
) -> Iterator[DataFrame]:
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield from _stream_dataframes(
                sel,
                conn,
                batch_size=batch_size,
                schema_overrides=schema_overrides,
                time_zone=time_zone,
                snake=snake,
                **kwargs,
            )
    else:
        for df in read_database(
            sel,
            cast(ConnectionOrCursor, engine_or_conn),
            iter_batches=True,
            batch_size=batch_size,
            schema_overrides=schema_overrides,
            **kwargs,
        ):
            yield _post_process(df, sel, time_zone=time_zone, snake=snake)


__all__ = [
    "ColumnToPolarsExprError",
    "insert_polars_dataframe",
    "PolarsDataFrameYieldsNoRowsError",
    "select_to_polars_dataframe",
]



from __future__ import annotations

from utilities.holoviews.holoviews import apply_opts, relabel_plot, save_plot

__all__ = ["apply_opts", "relabel_plot", "save_plot"]


try:
    from utilities.holoviews.xarray import (
        ArrayNameIsEmptyStringError,
        ArrayNameNotAStringError,
        plot_curve,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["ArrayNameIsEmptyStringError", "ArrayNameNotAStringError", "plot_curve"]



from __future__ import annotations

from pathlib import Path

from hypothesis import given
from hypothesis.strategies import DataObject, data
from pytest import raises

from utilities.git import (
    InvalidRepoError,
    get_branch_name,
    get_repo_name,
    get_repo_root,
)
from utilities.hypothesis import git_repos, text_ascii


class TestGetBranchName:
    @given(data=data(), branch=text_ascii(min_size=1))
    def test_main(self, *, data: DataObject, branch: str) -> None:
        path = data.draw(git_repos(branch=branch))
        result = get_branch_name(cwd=path)
        assert result == branch


class TestGetRepoName:
    def test_main(self) -> None:
        result = get_repo_name()
        expected = "python-utilities"
        assert result == expected


class TestGetRepoRoot:
    @given(path=git_repos())
    def test_main(self, *, path: Path) -> None:
        result = get_repo_root(cwd=path)
        assert result == path.resolve()

    def test_error(self, *, tmp_path: Path) -> None:
        with raises(InvalidRepoError):
            _ = get_repo_root(cwd=tmp_path)



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass
from functools import partial
from io import StringIO, TextIOWrapper
from multiprocessing import cpu_count
from typing import Any, Literal, TypeVar, cast

from pqdm import processes

from utilities.class_name import get_class_name
from utilities.sentinel import Sentinel, sentinel
from utilities.tqdm import _DEFAULTS as _TQDM_DEFAULTS
from utilities.tqdm import _get_total, _LockArgs, tqdm


@dataclass(frozen=True)
class _Defaults:
    parallelism: Literal["processes", "threads"] = "processes"
    n_jobs: int | None = None
    bounded: bool = False
    exception_behaviour: Literal["ignore", "immediate", "deferred"] = "immediate"


_PQDM_DEFAULTS = _Defaults()


_T = TypeVar("_T")


def pmap(
    func: Callable[..., _T],
    /,
    *iterables: Iterable[Any],
    parallelism: Literal["processes", "threads"] = _PQDM_DEFAULTS.parallelism,
    n_jobs: int | None = _PQDM_DEFAULTS.n_jobs,
    bounded: bool = _PQDM_DEFAULTS.bounded,
    exception_behaviour: Literal[
        "ignore", "immediate", "deferred"
    ] = _PQDM_DEFAULTS.exception_behaviour,
    desc: str | None | Sentinel = sentinel,
    total: float | None = _TQDM_DEFAULTS.total,
    leave: bool | None = _TQDM_DEFAULTS.leave,
    file: TextIOWrapper | StringIO | None = _TQDM_DEFAULTS.file,
    ncols: int | None = _TQDM_DEFAULTS.ncols,
    mininterval: float | None = _TQDM_DEFAULTS.mininterval,
    maxinterval: float | None = _TQDM_DEFAULTS.maxinterval,
    miniters: float | None = _TQDM_DEFAULTS.miniters,
    ascii: bool | str | None = _TQDM_DEFAULTS.ascii,  # noqa: A002
    unit: str | None = _TQDM_DEFAULTS.unit,
    unit_scale: bool | int | str | None = _TQDM_DEFAULTS.unit_scale,
    dynamic_ncols: bool | None = _TQDM_DEFAULTS.dynamic_ncols,
    smoothing: float | None = _TQDM_DEFAULTS.smoothing,
    bar_format: str | None = _TQDM_DEFAULTS.bar_format,
    initial: float | None = 0,
    position: int | None = _TQDM_DEFAULTS.position,
    postfix: Mapping[str, Any] | None = _TQDM_DEFAULTS.postfix,
    unit_divisor: float | None = _TQDM_DEFAULTS.unit_divisor,
    write_bytes: bool | None = _TQDM_DEFAULTS.write_bytes,
    lock_args: _LockArgs = _TQDM_DEFAULTS.lock_args,
    nrows: int | None = _TQDM_DEFAULTS.nrows,
    colour: str | None = _TQDM_DEFAULTS.colour,
    delay: float | None = _TQDM_DEFAULTS.delay,
    gui: bool | None = _TQDM_DEFAULTS.gui,
    **kwargs: Any,
) -> list[_T]:
    """Parallel map, powered by `pqdm`."""
    return pstarmap(
        func,
        zip(*iterables, strict=True),
        parallelism=parallelism,
        n_jobs=n_jobs,
        bounded=bounded,
        exception_behaviour=exception_behaviour,
        desc=desc,
        total=total,
        leave=leave,
        file=file,
        ncols=ncols,
        mininterval=mininterval,
        maxinterval=maxinterval,
        miniters=miniters,
        ascii=ascii,
        unit=unit,
        unit_scale=unit_scale,
        dynamic_ncols=dynamic_ncols,
        smoothing=smoothing,
        bar_format=bar_format,
        initial=initial,
        position=position,
        postfix=postfix,
        unit_divisor=unit_divisor,
        write_bytes=write_bytes,
        lock_args=lock_args,
        nrows=nrows,
        colour=colour,
        delay=delay,
        gui=gui,
        **kwargs,
    )


def pstarmap(
    func: Callable[..., _T],
    iterable: Iterable[tuple[Any, ...]],
    /,
    *,
    parallelism: Literal["processes", "threads"] = _PQDM_DEFAULTS.parallelism,
    n_jobs: int | None = _PQDM_DEFAULTS.n_jobs,
    bounded: bool = _PQDM_DEFAULTS.bounded,
    exception_behaviour: Literal[
        "ignore", "immediate", "deferred"
    ] = _PQDM_DEFAULTS.exception_behaviour,
    desc: str | None | Sentinel = sentinel,
    total: float | None = _TQDM_DEFAULTS.total,
    leave: bool | None = _TQDM_DEFAULTS.leave,
    file: TextIOWrapper | StringIO | None = _TQDM_DEFAULTS.file,
    ncols: int | None = _TQDM_DEFAULTS.ncols,
    mininterval: float | None = _TQDM_DEFAULTS.mininterval,
    maxinterval: float | None = _TQDM_DEFAULTS.maxinterval,
    miniters: float | None = _TQDM_DEFAULTS.miniters,
    ascii: bool | str | None = _TQDM_DEFAULTS.ascii,  # noqa: A002
    unit: str | None = _TQDM_DEFAULTS.unit,
    unit_scale: bool | int | str | None = _TQDM_DEFAULTS.unit_scale,
    dynamic_ncols: bool | None = _TQDM_DEFAULTS.dynamic_ncols,
    smoothing: float | None = _TQDM_DEFAULTS.smoothing,
    bar_format: str | None = _TQDM_DEFAULTS.bar_format,
    initial: float | None = 0,
    position: int | None = _TQDM_DEFAULTS.position,
    postfix: Mapping[str, Any] | None = _TQDM_DEFAULTS.postfix,
    unit_divisor: float | None = _TQDM_DEFAULTS.unit_divisor,
    write_bytes: bool | None = _TQDM_DEFAULTS.write_bytes,
    lock_args: tuple[Any, ...] | None = _TQDM_DEFAULTS.lock_args,
    nrows: int | None = _TQDM_DEFAULTS.nrows,
    colour: str | None = _TQDM_DEFAULTS.colour,
    delay: float | None = _TQDM_DEFAULTS.delay,
    gui: bool | None = _TQDM_DEFAULTS.gui,
    **kwargs: Any,
) -> list[_T]:
    """Parallel starmap, powered by `pqdm`."""
    n_jobs = _get_n_jobs(n_jobs)
    tqdm_class = cast(Any, tqdm)
    desc_kwargs = _get_desc(desc, func)
    total = _get_total(total, iterable)
    if parallelism == "processes":
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **desc_kwargs,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    else:
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **desc_kwargs,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    return list(result)


def _get_n_jobs(n_jobs: int | None, /) -> int:
    if (n_jobs is None) or (n_jobs <= 0):
        return cpu_count()  # pragma: no cover
    return n_jobs


def _get_desc(
    desc: str | None | Sentinel, func: Callable[..., Any], /
) -> dict[str, str]:
    if isinstance(desc, Sentinel):
        if isinstance(func, partial):
            return _get_desc(desc, func.func)
        try:
            desc_use = func.__name__
        except AttributeError:
            desc_use = get_class_name(func) if isinstance(func, object) else None
    else:
        desc_use = desc
    return {} if desc_use is None else {"desc": desc_use}


def _starmap_helper(func: Callable[..., _T], *args: Any) -> _T:
    return func(*args)


__all__ = ["pmap", "pstarmap"]



from __future__ import annotations

from utilities.polars.polars import (
    DataFrameColumnsError,
    DataFrameDTypesError,
    DataFrameHeightError,
    DataFrameMaxHeightError,
    DataFrameMinHeightError,
    DataFrameSchemaError,
    DataFrameShapeError,
    DataFrameSortedError,
    DataFrameUniqueError,
    DataFrameWidthError,
    EmptyDataFrameError,
    check_dataframe,
    join,
    set_first_row_as_columns,
)

__all__ = [
    "check_dataframe",
    "DataFrameColumnsError",
    "DataFrameDTypesError",
    "DataFrameHeightError",
    "DataFrameMaxHeightError",
    "DataFrameMinHeightError",
    "DataFrameSchemaError",
    "DataFrameShapeError",
    "DataFrameSortedError",
    "DataFrameUniqueError",
    "DataFrameWidthError",
    "EmptyDataFrameError",
    "join",
    "set_first_row_as_columns",
]


try:
    from utilities.polars.bs4 import MultipleTHRowsError, yield_tables
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["MultipleTHRowsError", "yield_tables"]



from __future__ import annotations

from pytest import mark, param
from streamlit.testing.v1 import AppTest

from utilities.os import temp_environ
from utilities.streamlit import _PASSWORD_CORRECT


class TestEnsureLoggedIn:
    @mark.parametrize("skip", [param(True), param(False)])
    @mark.parametrize("before_form", [param(True), param(False)])
    @mark.parametrize("after_form", [param(True), param(False)])
    def test_setup(self, *, skip: bool, before_form: bool, after_form: bool) -> None:
        def func() -> None:
            from os import getenv

            from streamlit import write

            from utilities.streamlit import ensure_logged_in

            skip = getenv("SKIP") == "1"
            if getenv("BEFORE_FORM") == "1":

                def _before_form() -> None:
                    _ = write("Before form")

                before_form = _before_form
            else:
                before_form = None
            if getenv("AFTER_FORM") == "1":

                def _after_form() -> None:
                    _ = write("after form")

                after_form = _after_form
            else:
                after_form = None
            ensure_logged_in(skip=skip, before_form=before_form, after_form=after_form)

        with temp_environ(
            SKIP="1" if skip else "0",
            BEFORE_FORM="1" if before_form else "0",
            AFTER_FORM="1" if after_form else "0",
        ):
            at = AppTest.from_function(func)
            _ = at.run()
            assert not at.exception

    @mark.parametrize("is_logged_in", [param(True), param(False)])
    def test_after_logged_in(self, *, is_logged_in: bool) -> None:
        def func() -> None:
            from utilities.streamlit import ensure_logged_in

            ensure_logged_in()

        at = AppTest.from_function(func)
        at.session_state[_PASSWORD_CORRECT] = is_logged_in
        _ = at.run()
        assert not at.exception
